@@ -32,16 +32,18 @@
  
 use std::ffi::{CStr, CString};
 use std::fmt;
-use std::net::{IpAddr, Ipv4Addr};
-use std::io::Error;
-use std::str::FromStr;
+use std::net::{IpAddr};
+use std::path::Path;
+use std::fs;
 
+
+use nix::NixPath;
 use nix::fcntl::{open, OFlag};
 use nix::sys::stat::Mode;
 
+use super::pf_tokenizer::PfListTokenizer;
 use crate::portable::net::{htonl, ntohl};
-use crate::{common::*, runtime_exception::*};
-use crate::{ cfor, map_runtime_error, runtime_error };
+use crate::{runtime_error, cfor, map_runtime_error, runtime_exception::*};
 
 
 const PF_TABLE_NAME_SIZE: usize = 32;
@@ -53,16 +55,16 @@ const IF_NAMESIZE: usize = 16;
 //libc is missing some constants for the FreeBSD
 const AI_NUMERICHOST: libc::c_int = 0x00000004;
 
-pub enum PfCmd<'pfcmd>
+pub enum PfCmd
 {
     /// Adds a host/hostname to table
-    Add{ host: &'pfcmd str }, 
+    Add{ hosts: Vec<String> }, 
 
     /// Removes a host/hostname from table
-    Delete{ host: &'pfcmd str }, 
+    Delete{ hosts: Vec<String> }, 
 
     /// Checks if host/hostname presents in table
-    Test{ host: &'pfcmd str },
+    Test{ hosts: Vec<String> },
 }
 
 
@@ -122,30 +124,62 @@ struct pf_addr
     pfa: pf_addr_union,
 }
 
-impl pf_addr
+impl From<libc::in_addr> for pf_addr
 {
-    fn new_empty() -> Self
+    fn from(addr: libc::in_addr) -> Self 
     {
-        return Self{
-            pfa: pf_addr_union{ addr32: [0, 0, 0, 0] }
-        };
-    }
+        // zero memory first
+        let mut s: Self = unsafe{ std::mem::zeroed() };
 
-    fn new_ipv4(octets: [u8; 4]) -> Self
+        s.pfa.v4 = addr;
+        
+        return s;
+    }
+}
+
+impl From<libc::in6_addr> for pf_addr
+{
+    fn from(addr: libc::in6_addr) -> Self 
+    {
+        // zero memory first
+        let mut s: Self = unsafe{ std::mem::zeroed() };
+
+        s.pfa.v6 = addr;
+        
+        return s;
+    }
+}
+
+/// For IPv4
+impl From<[u8; 4]> for pf_addr
+{
+    fn from(octets: [u8; 4]) -> Self 
     {
         let mut s: Self = unsafe{ std::mem::zeroed() };
         s.pfa.v4.s_addr = u32::from_le_bytes(octets);
 
         return s;
     }
+}
 
-    fn new_ipv6(octets: [u8; 16]) -> Self 
+/// For IPv6
+impl From<[u8; 16]> for pf_addr
+{
+    fn from(octets: [u8; 16]) -> Self 
     {
         let mut s: Self = unsafe{ std::mem::zeroed() };
         s.pfa.v6.s6_addr = octets;
 
         return s;
     }
+}
+
+impl pf_addr
+{
+    fn new_empty() -> Self
+    {
+        return unsafe{ std::mem::zeroed() };
+    }
 
     unsafe fn pfctl_addrprefix<'addr>(&mut self, addr: &'addr str) -> PfResult<()>
     {
@@ -427,7 +461,7 @@ impl node_host
 {
 
     // https://github.com/freebsd/freebsd-src/blob/27ab791a55191c0b6503391d411303b042b41047/sbin/pfctl/pfctl_parser.c :1648
-    unsafe fn host_v4v6<'pfcmd>(s: &'pfcmd str, maskv4: libc::c_int, maskv6: libc::c_int) -> PfResult<Option<Self>>
+    unsafe fn host_v4v6<'pfcmd>(s: &'pfcmd str, maskv4: libc::c_int, maskv6: libc::c_int) -> PfResult<Option<Vec<Self>>>
     {
         let ipaddr: IpAddr = 
             match s.parse()
@@ -436,11 +470,13 @@ impl node_host
                 Err(e) => return Ok(None),
             };
 
+        let mut node_host_list = Vec::new();
+
         match ipaddr
         {
             IpAddr::V4(ipv4) =>
             {
-                let addr = pf_addr::new_ipv4(ipv4.octets());
+                let addr = pf_addr::from(ipv4.octets());
 
                 let mut addr_wrap: pf_addr_wrap = pf_addr_wrap::set_ipmask(PfAddr::PF_ADDR_ADDRMASK, maskv4, addr);
 
@@ -458,8 +494,11 @@ impl node_host
                         ifname: std::ptr::null(),
                         ifa_flags: 0,
                     };
+
+                // add to list i.e h->next = NULL; h->tail = h;
+                node_host_list.push(h);
             
-                return Ok(Some( h ));
+                return Ok(Some( node_host_list ));
             },
             IpAddr::V6(_ipv6) => // https://github.com/freebsd/freebsd-src/blob/27ab791a55191c0b6503391d411303b042b41047/sbin/pfctl/pfctl_parser.c#L1677
             {
@@ -480,10 +519,12 @@ impl node_host
         
                 if ret_ga > 0
                 {
-                    runtime_error!("getaddrinfo: {}", CStr::from_ptr(libc::gai_strerror(ret_ga)).to_string_lossy());
+                    //runtime_error!("getaddrinfo: {}", CStr::from_ptr(libc::gai_strerror(ret_ga)).to_string_lossy());
+                    return Ok(None);
                 }
                 else if res == std::ptr::null_mut()
                 {
+                    // assertion
                     runtime_error!("getaddrinfo: returned NULL");
                 }
 
@@ -500,9 +541,9 @@ impl node_host
                 );*/
 
                 //h->ifindex = ((struct sockaddr_in6 *)res->ai_addr)->sin6_scope_id;
-                // h.ifindex = (*s_in6).sin6_scope_id;
+            // h.ifindex = (*s_in6).sin6_scope_id;
                 let s_in6 = (*res).ai_addr as *const libc::sockaddr_in6;
-                let addr = pf_addr::new_ipv6((*s_in6).sin6_addr.s6_addr);
+                let addr = pf_addr::from((*s_in6).sin6_addr);
                 let addr_wrap: pf_addr_wrap = pf_addr_wrap::set_ipmask(PfAddr::PF_ADDR_ADDRMASK, maskv6, addr);
 
                 let h = 
@@ -518,14 +559,163 @@ impl node_host
                         ifa_flags: 0,
                     };
 
+                // add to list i.e h->next = NULL; h->tail = h;
+                node_host_list.push(h);
+
                 libc::freeaddrinfo(res);
 
-                return Ok(Some( h ));
+                return Ok(Some( node_host_list ));
             },
         }
     }
 
-    unsafe fn host<'pfcmd>(s: &'pfcmd str) -> PfResult<Self>
+    unsafe fn host_dns<'pfcmd>(s: &'pfcmd str, maskv4: libc::c_int, maskv6: libc::c_int) -> PfResult<Option<Vec<Self>>>
+    {
+        let mut noalias: libc::c_int = 0;
+
+        let ps =
+            if let Some(pos) = s.chars().position(|c| c == ':') // p = strrchr(ps, ':')) != NULL
+            {
+                match s.char_indices().nth(pos) 
+                {
+                    None => runtime_error!("can not get char indeces at: '{}' for: '{}'", pos, s),
+                    Some((idx, _)) => 
+                    {
+                        if &s[idx..] == ":0" // && !strcmp(p, ":0") i.e strcmp(p, ":0") == 0
+                        {
+                            noalias = 1;
+
+                            &s[..idx]
+                        }
+                        else
+                        {
+                            s
+                        }
+                    },
+                }
+            }
+            else
+            {
+                s
+            };
+
+            let mut res: *mut libc::addrinfo = std::ptr::null_mut();
+            let mut hints: libc::addrinfo = std::mem::zeroed();
+
+            hints.ai_family = libc::PF_UNSPEC;
+            hints.ai_socktype = libc::SOCK_STREAM; // DUMMY
+
+            let c_ps = CString::new(s.to_string()).map_err(|e| map_runtime_error!("{}", e))?;
+
+            // getaddrinfo(s, "0", &hints, &res
+            let ret_ga = 
+                libc::getaddrinfo(
+                    c_ps.as_ptr(), 
+                    std::ptr::null(), 
+                    &hints as *const _ as *const libc::addrinfo, 
+                    &mut res
+                );
+    
+            if ret_ga > 0
+            {
+                //runtime_error!("getaddrinfo: {}", CStr::from_ptr(libc::gai_strerror(ret_ga)).to_string_lossy());
+                return Ok(None);
+            }
+            else if res == std::ptr::null_mut()
+            {
+                runtime_error!("getaddrinfo: returned NULL");
+            }
+
+            let mut got4: libc::c_int = 0;
+            let mut got6: libc::c_int = 0;
+            let mut node_host_list: Vec<node_host> = Vec::new();
+
+            cfor!(let mut p_resp = res; p_resp != std::ptr::null_mut(); p_resp = (*p_resp).ai_next;
+                {
+                    let resp = *p_resp;
+
+                    if resp.ai_family != libc::AF_INET &&
+                        resp.ai_family != libc::AF_INET6
+                    {
+                        continue;
+                    }
+                
+
+                    if noalias > 0
+                    {
+                        if resp.ai_family == libc::AF_INET
+                        {
+                            if got4 > 0
+                            {
+                                continue;
+                            }
+
+                            got4 = 1;
+                        }
+                        else
+                        {
+                            if got6 > 0
+                            {
+                                continue;
+                            }
+
+                            got6 = 1;
+                        }
+                    } // if noalias
+
+                    let mut ifindex: u32 = 0;
+
+                    let addr_wrap: pf_addr_wrap = 
+                        if resp.ai_family == libc::AF_INET
+                        {
+                            let s_in4: libc::sockaddr_in = *(resp.ai_addr as *const libc::sockaddr_in);
+                            let addr = pf_addr::from(s_in4.sin_addr);
+
+                            pf_addr_wrap::set_ipmask(PfAddr::PF_ADDR_ADDRMASK, maskv4, addr)
+                        }
+                        else 
+                        {
+                            let s_in6 = resp.ai_addr as *const libc::sockaddr_in6;
+
+                            ifindex = (*s_in6).sin6_scope_id;
+
+                            let addr = pf_addr::from((*s_in6).sin6_addr);
+
+                            pf_addr_wrap::set_ipmask(PfAddr::PF_ADDR_ADDRMASK, maskv6, addr)
+                        };
+
+                    let n: Self = 
+                        Self
+                        {
+                            addr: addr_wrap,
+                            bcast: pf_addr::new_empty(),
+                            peer: pf_addr::new_empty(),
+                            af: resp.ai_family as libc::sa_family_t,
+                            not: 0,
+                            ifindex: ifindex,
+                            ifname: std::ptr::null(),
+                            ifa_flags: 0,
+                        };
+
+                    node_host_list.push(n);
+                }
+            );
+
+            libc::freeaddrinfo(res);
+
+            return Ok(
+                if node_host_list.is_empty() == true
+                {
+                    None
+                }
+                else
+                {
+                    Some(node_host_list)
+                }
+            );
+    }
+
+    unsafe fn host<'pfcmd>(s: &'pfcmd str) -> PfResult<Vec<Self>>
     {
         let mut v4mask: libc::c_int = 32;
         let mut v6mask: libc::c_int = 128;
@@ -586,13 +776,17 @@ impl node_host
         {
             return Ok(r);
         }
-        /*else if let Some(r) = host_dns(str_ip, v4mask, v6mask)
+        /*else if let Some(r) => Self::host_if(str_ip, mask)
         {
-            //todo
+
         }*/
+        else if let Some(r) = Self::host_dns(str_ip, v4mask, v6mask)?
+        {
+            return Ok(r);
+        }
         else
         {
-            runtime_error!("invalid address: {} not IP or DNS", str_ip);
+            runtime_error!("invalid address: {} not IP or DNS or IF", str_ip);
         }
 
         
@@ -627,7 +821,7 @@ impl pfr_addr
         let mut bits: libc::c_int = 0;
 
         let pfra_u = 
-            match n.af as i32             // switch (n->af) { 
+            match n.af as i32             // switch (n->af) {
             {
                 libc::AF_INET => 
                 {
@@ -705,7 +899,7 @@ struct pfioc_state_kill
 
 impl pfioc_state_kill
 {
-    unsafe fn new<H: AsRef<str>>(fd: libc::c_int, is_test: bool, src: H, dest: Option<H>) -> PfResult<i32>
+    unsafe fn new<H: AsRef<str>>(fd: libc::c_int, is_test: bool, src: H, dest: Option<H>) -> PfResult<()>
     {
         let host = src.as_ref();
 
@@ -740,7 +934,7 @@ impl pfioc_state_kill
             libc::memset(&mut last_dst as *mut _ as *mut libc::c_void, 0xff, std::mem::size_of::<libc::sockaddr>());
 
         let mut sources: libc::c_int = 0;
-        let mut killed: i32 = 0;
+        let mut killed: u32 = 0;
         let mut dests: libc::c_int = 0;
 
         
@@ -859,7 +1053,7 @@ impl pfioc_state_kill
 
                             if is_test == false
                             {
-                                if libc::ioctl(fd, IoWr::DIOCKILLSTATES, &mut psk as *mut _, std::ptr::null::<libc::c_void>()) != 0
+                                if libc::ioctl(fd, IoCtl::DIOCKILLSTATES, &mut psk as *mut _, std::ptr::null::<libc::c_void>()) != 0
                                 {
                                     //pfr_report_error(tbl, &io, "add table");
                                     libc::freeaddrinfo(res);
@@ -870,10 +1064,10 @@ impl pfioc_state_kill
                             }
                             else
                             {
-                                println!("Fake ioctl DIOCKILLSTATES: {}\n{:?}", IoWr::DIOCKILLSTATES, psk);
+                                println!("Fake ioctl DIOCKILLSTATES: {}\n{:?}", IoCtl::DIOCKILLSTATES, psk);
                             }
 
-                            killed += psk.psk_killed as i32;
+                            killed += psk.psk_killed;
 
                         }
                     );
@@ -884,7 +1078,7 @@ impl pfioc_state_kill
                 {
                     if is_test == false
                     {
-                        if libc::ioctl(fd, IoWr::DIOCKILLSTATES, &mut psk as *mut _, std::ptr::null::<libc::c_void>()) != 0
+                        if libc::ioctl(fd, IoCtl::DIOCKILLSTATES, &mut psk as *mut _, std::ptr::null::<libc::c_void>()) != 0
                         {
                             //pfr_report_error(tbl, &io, "add table");
                             libc::freeaddrinfo(res);
@@ -894,10 +1088,10 @@ impl pfioc_state_kill
                     }
                     else
                     {
-                        println!("Fake ioctl DIOCKILLSTATES: {}\n{:?}", IoWr::DIOCKILLSTATES, psk);
+                        println!("Fake ioctl DIOCKILLSTATES: {}\n{:?}", IoCtl::DIOCKILLSTATES, psk);
                     }
 
-                    killed += psk.psk_killed as i32;
+                    killed += psk.psk_killed;
                 }
             }
         );
@@ -907,7 +1101,7 @@ impl pfioc_state_kill
         #[cfg(feature = "log_to_stdout")]
         println!("killed {} states from {} sources and {} destinations", killed, sources, dests);
 
-        return Ok(killed);
+        return Ok(());
     }
 }
 
@@ -955,8 +1149,8 @@ impl pfioc_table
     }
 }
 
-struct IoWr{}
-impl IoWr
+struct IoCtl{}
+impl IoCtl
 {
     /// no parameters
     const IOC_VOID: libc::c_ulong = 0x20000000;
@@ -1057,7 +1251,7 @@ impl pfr_table
 
         if test == false
         {
-            if libc::ioctl(fd, IoWr::DIOCRADDTABLES, &mut io as *mut _, std::ptr::null::<libc::c_void>()) != 0
+            if libc::ioctl(fd, IoCtl::DIOCRADDTABLES, &mut io as *mut _, std::ptr::null::<libc::c_void>()) != 0
             {
                 //pfr_report_error(tbl, &io, "add table");
                 runtime_error!("ioctl DIOCRADDTABLES failed");
@@ -1097,7 +1291,7 @@ impl pfr_table
         
         if test == false
         {
-            if libc::ioctl(fd, IoWr::DIOCRADDADDRS, &mut io as *mut _, std::ptr::null::<libc::c_void>()) != 0
+            if libc::ioctl(fd, IoCtl::DIOCRADDADDRS, &mut io as *mut _, std::ptr::null::<libc::c_void>()) != 0
             {
                 //pfr_report_error(tbl, &io, "add table");
                 runtime_error!("ioctl DIOCRADDADDRS failed");
@@ -1128,7 +1322,7 @@ impl pfr_table
                 flags
             );    
         
-        if libc::ioctl(fd, IoWr::DIOCRDELADDRS, &mut io as *mut _, std::ptr::null::<libc::c_void>()) != 0
+        if libc::ioctl(fd, IoCtl::DIOCRDELADDRS, &mut io as *mut _, std::ptr::null::<libc::c_void>()) != 0
         {
             //pfr_report_error(tbl, &io, "add table");
             runtime_error!("ioctl DIOCRDELADDRS failed");
@@ -1154,7 +1348,7 @@ impl pfr_table
                 flags
             );    
         
-        if libc::ioctl(fd, IoWr::DIOCRTSTADDRS, &mut io as *mut _, std::ptr::null::<libc::c_void>()) != 0
+        if libc::ioctl(fd, IoCtl::DIOCRTSTADDRS, &mut io as *mut _, std::ptr::null::<libc::c_void>()) != 0
         {
             //pfr_report_error(tbl, &io, "add table");
             runtime_error!("ioctl DIOCRDELADDRS failed");
@@ -1174,9 +1368,9 @@ impl pfr_table
 
         self.pfr_add_tables(fd, 1, &mut nadd, flags, test)?;
 
+        #[cfg(feature = "log_to_stdout")]
         if nadd > 0
         {
-            #[cfg(feature = "log_to_stdout")]
             println!("{} table created", nadd);
         }
 
@@ -1185,7 +1379,7 @@ impl pfr_table
         return Ok(());
     }
 
-    unsafe fn add_addrs(&mut self, fd: libc::c_int, b: PfrBuffer, test: bool) -> PfResult<i32>
+    unsafe fn add_addrs(&mut self, fd: libc::c_int, b: PfrBuffer, test: bool) -> PfResult<()>
     {
         let mut nadd: libc::c_int = 0;
         let flags: libc::c_int = 0;
@@ -1196,25 +1390,24 @@ impl pfr_table
         #[cfg(feature = "log_to_stdout")]
         println!("{}/{} addresses added", nadd, pfrb_size);
 
-        return Ok(nadd);
+        return Ok(());
     }
 
-    unsafe fn del_addrs(&mut self, fd: libc::c_int, b: PfrBuffer) -> PfResult<i32>
+    unsafe fn del_addrs(&mut self, fd: libc::c_int, b: PfrBuffer) -> PfResult<()>
     {
-        let mut ndel: libc::c_int = 0;
+        let mut nadd: libc::c_int = 0;
         let flags: libc::c_int = 0;
 
         let pfrb_size = b.pfrb_caddr.len();
-        self.pfr_del_addrs(fd, b, &mut ndel, flags)?;
+        self.pfr_del_addrs(fd, b, &mut nadd, flags)?;
 
         #[cfg(feature = "log_to_stdout")]
-        println!("[driver: pf] {}/{} addresses deleted", ndel, pfrb_size);
+        println!("{}/{} addresses deleted", nadd, pfrb_size);
 
-        return Ok(ndel);
+        return Ok(());
     }
 
-    /// returns amount of matched IP
-    unsafe fn tst_addrs(&mut self, fd: libc::c_int, b: PfrBuffer) -> PfResult<i32>
+    unsafe fn tst_addrs(&mut self, fd: libc::c_int, b: PfrBuffer) -> PfResult<bool>
     {
         let mut nmatch = 0;
         let flags: libc::c_int = 0;
@@ -1223,9 +1416,9 @@ impl pfr_table
         self.pfr_tst_addrs(fd, b, &mut nmatch, flags)?;
 
         #[cfg(feature = "log_to_stdout")]
-        println!("[driver: pf] {}/{} addresses test", nmatch, pfrb_size);
+        println!("{}/{} addresses test", nmatch, pfrb_size);
 
-        return Ok(nmatch);
+        return Ok(nmatch as usize == pfrb_size);
     }
 
 }
@@ -1291,10 +1484,10 @@ impl PfrBuffer
     }
 
     //https://github.com/freebsd/freebsd-src/blob/373ffc62c158e52cde86a5b934ab4a51307f9f2e/sbin/pfctl/pfctl_table.c : 416
-    unsafe fn load_addr<'pfcmd>(&mut self, s: &'pfcmd str, is_test: libc::c_int) -> PfResult<()>
+    unsafe fn load_addr_from_str<H: AsRef<str>>(&mut self, s: H, is_test: libc::c_int) -> PfResult<()>
     {
         //while (argc--)
-        self.append_addr(s, is_test)?;
+        self.append_addr(s.as_ref(), is_test)?;
 
         //todo (from file if needed)
         /*if (pfr_buf_load(b, file, nonetwork, append_addr)) 
@@ -1306,6 +1499,81 @@ impl PfrBuffer
         return Ok(());
     }
 
+    
+    unsafe fn load_addr_from_file<P: AsRef<Path>>(&mut self, path: P, is_test: libc::c_int) -> PfResult<()>
+    {
+
+        let s_path = path.as_ref();
+
+        if s_path.as_os_str().len() == 0
+        {
+            #[cfg(feature = "log_to_stdout")]
+            println!("load_addr_from_file() path is empty!");
+
+            return Ok(());
+        }
+        
+        let mut stream: Box<dyn std::io::Read + 'static> = 
+            if s_path.as_os_str() == "-"
+            {
+                // read from stdin
+                Box::new(std::io::stdin())
+            }
+            else
+            {   
+                if s_path.exists() == true
+                {
+                    runtime_error!("path does not exist: {}", s_path.display());
+                }
+                else if s_path.is_dir() == true
+                {
+                    runtime_error!("path is dir: {}", s_path.display());
+                }
+
+                Box::new(fs::File::open(s_path).map_err(|e| map_runtime_error!("{}", e))?)
+            };
+
+        //read from stream all
+        let mut file_buf = String::new();
+        stream.read_to_string(&mut file_buf).map_err(|e| map_runtime_error!("{}", e))?;
+
+        let token = PfListTokenizer::from_str(&file_buf)?;
+        
+        if token.is_none() == true
+        {
+            return Ok(());
+        } 
+
+        let mut token = token.unwrap();
+
+        loop
+        {
+            let tok = token.read_next()?;
+
+            match tok
+            {
+                Some(s) => 
+                {
+                    self.append_addr(s, is_test)?;
+                },
+                None => break,
+            }
+        }
+
+        //while (argc--)
+        
+
+        //todo (from file if needed)
+        /*if (pfr_buf_load(b, file, nonetwork, append_addr)) 
+        {
+            warn("cannot load %s", file);
+            return (-1);
+        }*/
+
+        return Ok(());
+    }
+    
+
     // https://github.com/freebsd/freebsd-src/blob/373ffc62c158e52cde86a5b934ab4a51307f9f2e/sbin/pfctl/pfctl_radix.c#L457
     fn pfr_buf_add(&mut self, addr: pfr_addr) -> PfResult<()>
     {
@@ -1321,24 +1589,28 @@ impl PfrBuffer
         return Ok(());
     }
 
-    unsafe fn append_addr_host(&mut self, mut n: node_host, is_test: libc::c_int, not: libc::c_int) -> PfResult<()>
+    unsafe fn append_addr_host(&mut self, vn: Vec<node_host>, is_test: libc::c_int, not: libc::c_int) -> PfResult<()>
     {
-        let addr = pfr_addr::new(n, is_test, not)?;
+        for n in vn
+        {
+            let addr = pfr_addr::new(n, is_test, not)?;
 
-        self.pfr_buf_add(addr)?;
+            self.pfr_buf_add(addr)?;
+        }
 
         return Ok(());
     }
 }
 
-// todo move to separate file
-pub trait NativeActionDriver
+/*fn pfctl_fopen(file: &Path) -> PfResult<fs::File>
 {
-    fn kill_state<H: AsRef<str>>(&self, host: H) -> PfResult<()>;
-    fn block<H: AsRef<str>>(&self, host: H) -> PfResult<()>;
-    fn unblock<H: AsRef<str>>(&self, host: H) -> PfResult<()>;
-    fn check<H: AsRef<str>>(&self, host: H) -> PfResult<bool>;
-}
+    if file.is_dir() == true
+    {
+        runtime_error!("path is dir: {}", file.display());
+    }
+
+    return Ok(fs::File::open(file).map_err(|e| map_runtime_error!("{}", e))?);
+}*/
 
 
 pub struct Pf
@@ -1351,8 +1623,6 @@ impl Pf
 {
     const DEFAULT_PF_DEVICE: &'static str = "/dev/pf";
 
-    /// Creates new instance which if `test` is set to false attempts to open
-    /// a fd to /dev/pf. Requires root permission.
     pub fn new(test: bool) -> PfResult<Self>
     {
         let fd = 
@@ -1369,15 +1639,14 @@ impl Pf
         return Ok(Self{ fd: fd, test: test });
     }
 
-    /// Kills the state form the state table by source IP or with destination IP
-    pub unsafe fn pfctl_kill_state<H: AsRef<str>>(&self, src: H, dest: Option<H>) -> PfResult<i32>
+    // internal functions to be called from public interface
+    // https://github.com/freebsd/freebsd-src/blob/373ffc62c158e52cde86a5b934ab4a51307f9f2e/sbin/pfctl/pfctl.c#L638
+    unsafe fn pfctl_kill_state<H: AsRef<str>>(&self, src: H, dest: Option<H>) -> PfResult<()>
     {
         return pfioc_state_kill::new(self.fd, self.test, src, dest);
     }
 
-    /// Operations on records of a specific table: add, test, delete.
-    /// Requires: table_name - a name of the table
-    pub unsafe fn pfctl_table<T: AsRef<str>>(&self, table_name: T, cmd: PfCmd) -> PfResult<i32>
+    unsafe fn pfctl_table<T: AsRef<str>>(&self, table_name: T, cmd: PfCmd) -> PfResult<()>
     {
         let tname = table_name.as_ref();
 
@@ -1387,55 +1656,62 @@ impl Pf
 
         if tname.len() >= PF_TABLE_NAME_SIZE
         {
-            runtime_error!("table name: '{}' is too long!", tname);
+            runtime_error!("pf [action]: table name: '{}' is too long!", tname);
         }
 
         // verify that onlu ASCII chars are in tname
         if tname.chars().all(|c| c.is_ascii()) == false
         {
-            runtime_error!("non ascii character in table_name: {}", tname);
+            runtime_error!("pf [action]: non ascii character in table_name: {}", tname);
         }
 
         //table.pfrt_name.copy_from_slice(tname.as_bytes());
 
         // https://github.com/freebsd/freebsd-src/blob/373ffc62c158e52cde86a5b934ab4a51307f9f2e/sbin/pfctl/pfctl_table.c :201
-
-        let ret = 
-            match cmd
+        match cmd
+        {
+            PfCmd::Add{ hosts } =>
             {
-                PfCmd::Add{ host } =>
+                let mut b: PfrBuffer = PfrBuffer::new(Pfrb::PFRB_ADDRS);
+                //b.pfrb_type = Pfrb::PFRB_ADDRS as libc::c_int; //C b.pfrb_type = PFRB_ADDRS;
+
+                for host in hosts
                 {
-                    let mut b: PfrBuffer = PfrBuffer::new(Pfrb::PFRB_ADDRS);
-                    //b.pfrb_type = Pfrb::PFRB_ADDRS as libc::c_int; //C b.pfrb_type = PFRB_ADDRS;
+                    b.load_addr_from_str(host, 0)?; //C if (load_addr(&b, argc, argv, file, 0))
+                }
 
-                    b.load_addr(host, 0)?; //C if (load_addr(&b, argc, argv, file, 0))
+                table.create_table(self.fd, self.test)?;
 
-                    table.create_table(self.fd, self.test)?;
+                table.add_addrs(self.fd, b, self.test)?;
+            },
+            PfCmd::Delete{ hosts } => 
+            {
+                let mut b: PfrBuffer = PfrBuffer::new(Pfrb::PFRB_ADDRS);
+                //b.pfrb_type = Pfrb::PFRB_ADDRS as libc::c_int; //C b.pfrb_type = PFRB_ADDRS;
 
-                    table.add_addrs(self.fd, b, self.test)?
-                },
-                PfCmd::Delete{ host } => 
+                for host in hosts
                 {
-                    let mut b: PfrBuffer = PfrBuffer::new(Pfrb::PFRB_ADDRS);
-                    //b.pfrb_type = Pfrb::PFRB_ADDRS as libc::c_int; //C b.pfrb_type = PFRB_ADDRS;
+                    b.load_addr_from_str(host, 0)?; //C if (load_addr(&b, argc, argv, file, 0))
+                }
 
-                    b.load_addr(host, 0)?; //C if (load_addr(&b, argc, argv, file, 0))
+                table.del_addrs(self.fd, b)?;
+            },
+            PfCmd::Test{ hosts } => 
+            {
+                let mut b: PfrBuffer = PfrBuffer::new(Pfrb::PFRB_ADDRS);
+                //b.pfrb_type = Pfrb::PFRB_ADDRS as libc::c_int; //C b.pfrb_type = PFRB_ADDRS;
+                //b2.pfrb_type = Pfrb::PFRB_ADDRS as libc::c_int; //C b.pfrb_type = PFRB_ADDRS;
 
-                    table.del_addrs(self.fd, b)?
-                },
-                PfCmd::Test{ host } => 
+                for host in hosts
                 {
-                    let mut b: PfrBuffer = PfrBuffer::new(Pfrb::PFRB_ADDRS);
-                    //b.pfrb_type = Pfrb::PFRB_ADDRS as libc::c_int; //C b.pfrb_type = PFRB_ADDRS;
-                    //b2.pfrb_type = Pfrb::PFRB_ADDRS as libc::c_int; //C b.pfrb_type = PFRB_ADDRS;
+                    b.load_addr_from_str(host, 1)?; //C if (load_addr(&b, argc, argv, file, 1))
+                }
 
-                    b.load_addr(host, 1)?; //C if (load_addr(&b, argc, argv, file, 1))
+                table.tst_addrs(self.fd, b)?;
+            },
+        }
 
-                    table.tst_addrs(self.fd, b)?
-                },
-            };
-
-        return Ok(ret);
+        return Ok(());
     }
 
     
@@ -1446,7 +1722,7 @@ fn test_real_add()
 {
     let pf = Pf::new(false).unwrap();
 
-    unsafe { pf.pfctl_table("test", PfCmd::Add{ host: "192.168.2.1" }).unwrap() };
+    unsafe { pf.pfctl_table("test", PfCmd::Add{ hosts: vec!["192.168.2.1".to_string()] }).unwrap() };
 }
 
 #[test]
@@ -1454,7 +1730,7 @@ fn test_real_check()
 {
     let pf = Pf::new(false).unwrap();
 
-    unsafe { pf.pfctl_table("test", PfCmd::Test{ host: "192.168.2.1" }).unwrap() };
+    unsafe { pf.pfctl_table("test", PfCmd::Test{ hosts: vec!["192.168.2.1".to_string()] }).unwrap() };
 }
 
 #[test]
@@ -1462,7 +1738,7 @@ fn test_real_del()
 {
     let pf = Pf::new(false).unwrap();
 
-    unsafe { pf.pfctl_table("test", PfCmd::Delete{ host: "192.168.2.1" }).unwrap() };
+    unsafe { pf.pfctl_table("test", PfCmd::Delete{ hosts: vec!["192.168.2.1".to_string()] }).unwrap() };
 }
 
 #[test]
@@ -1486,7 +1762,7 @@ fn fake_test_add()
 {
     let pf = Pf::new(true).unwrap();
 
-    unsafe { pf.pfctl_table("test", PfCmd::Add{ host: "192.168.2.1" }).unwrap() };
+    unsafe { pf.pfctl_table("test", PfCmd::Add{ hosts: vec!["192.168.2.1".to_string()] }).unwrap() };
 }
 
 #[test]
@@ -1494,7 +1770,7 @@ fn fake_test2_add()
 {
     let pf = Pf::new(true).unwrap();
 
-    unsafe { pf.pfctl_table("test", PfCmd::Add{ host: "192.168.2.0/24" }).unwrap() };
+    unsafe { pf.pfctl_table("test", PfCmd::Add{ hosts: vec!["192.168.2.0/24".to_string()] }).unwrap() };
 }
 
 #[test]
@@ -1588,11 +1864,13 @@ psk.psk_src.addr.iflags size: 1
 #[test]
 fn test_unmask()
 {
+    use std::net::Ipv4Addr;
+
     fn create(ip: &'static str) -> libc::c_int
     {
         let ipv4: Ipv4Addr = ip.parse().unwrap();
 
-        let ip = pf_addr::new_ipv4(ipv4.octets());
+        let ip = pf_addr::from(ipv4.octets());
 
         return unsafe {ip.unmask(0)};
     }
@@ -1614,8 +1892,9 @@ fn test_set_ipmask()
 {
     unsafe
     {
-        let h = node_host::host("192.168.2.1").unwrap();
+        let vh = node_host::host("192.168.2.1").unwrap();
 
+        let h = &vh[0];
         assert_eq!(h.af, 2);
         assert_eq!(h.addr.v.a.addr.pfa.addr32[0], 16951488);
         assert_eq!(h.addr.v.a.mask.pfa.addr32[0], 4294967295);
@@ -1627,7 +1906,9 @@ fn test_set_ipmask_cidr()
 {
     unsafe
     {
-        let h = node_host::host("192.168.2.0/24").unwrap();
+        let vh = node_host::host("192.168.2.0/24").unwrap();
+
+        let h = &vh[0];
 
         assert_eq!(h.af, 2);
         assert_eq!(h.addr.v.a.addr.pfa.addr32[0], 174272);
@@ -1635,4 +1916,3 @@ fn test_set_ipmask_cidr()
     }
 }
 
-
